import {
  useAuthentication,
  useAuthCache,
  useAppDispatch,
  useSessionCache,
  useQuerybuilderHash,
  useMLEnabledHash,
  useQuerybuilderSettings,
  readInSchemaFromBackend,
  setNewGraphQueryResult,
  useQuerybuilder,
  useVisualization,
  wsSchemaRequest,
  wsSchemaSubscription,
  useQuerybuilderAttributesShown,
  wsSchemaStatsRequest,
} from '@/lib/data-access';
import { Broker, wsQuerySubscription, wsQueryTranslationSubscription } from '@/lib/data-access/broker';
import { addInfo } from '@/lib/data-access/store/configSlice';
import { setMLResult } from '@/lib/data-access/store/mlSlice';
import { setQueryText, setQuerybuilderNodes } from '@/lib/data-access/store/querybuilderSlice';
import { useEffect } from 'react';
import {
  wsGetState,
  wsGetStates,
  wsUpdateState,
  nilUUID,
  wsGetStatesSubscription,
  wsDeleteStateSubscription,
  wsGetStateSubscription,
  wsTestSaveStateConnectionSubscription,
  wsStateGetPolicy,
} from '../broker/wsState';
import {
  addSaveState,
  testedSaveState,
  selectSaveState,
  updateSaveStateList,
  updateSelectedSaveState,
  setFetchingSaveStates,
  setStateAuthorization,
} from '../store/sessionSlice';
import { URLParams, getParam } from './url';
import { VisState, setVisualizationState } from '../store/visualizationSlice';
import { isEqual } from 'lodash-es';
import { setSchemaAttributeDimensions, setSchemaAttributeInformation, setSchemaLoading } from '../store/schemaSlice';
import { addError } from '@/lib/data-access/store/configSlice';
import { unSelect } from '../store/interactionSlice';
import { wsReconnectSubscription, wsUserGetPolicy } from '../broker/wsUser';
import { authorized, setSessionID } from '../store/authSlice';
import { queryingBackend } from '../store/graphQueryResultSlice';
import {
  SaveState,
  SchemaGraphStats,
  allMLTypes,
  LinkPredictionInstance,
  GraphQueryTranslationResultMessage,
  wsReturnKey,
} from 'ts-common';

export const EventBus = (props: { onRunQuery: () => void; onAuthorized: () => void }) => {
  const { login } = useAuthentication();
  const auth = useAuthCache();
  const dispatch = useAppDispatch();
  const session = useSessionCache();
  const queryHash = useQuerybuilderHash();
  const queryBuilder = useQuerybuilder();
  const attributeShown = useQuerybuilderAttributesShown();
  const mlHash = useMLEnabledHash();
  const visState = useVisualization();
  const queryBuilderSettings = useQuerybuilderSettings();

  function loadSaveState(saveStateID: string | undefined, saveStates: Record<string, SaveState>) {
    if (saveStateID && saveStates && saveStateID in saveStates) {
      console.debug('Setting state fetched from database', saveStateID, saveStates);
      const state = saveStates[saveStateID];
      if (state) {
        if (state.queries && state.queries.length > 0) {
          dispatch(setQuerybuilderNodes(state.queries[0]));
        }
        dispatch(
          setVisualizationState(
            Object.keys(state.visualizations).length !== 0 // should only occur in mock data
              ? (state.visualizations as VisState)
              : {
                  activeVisualizationIndex: -1,
                  openVisualizationArray: [],
                },
          ),
        );
      }
    }
  }

  useEffect(() => {
    const unsubs: (() => void)[] = [];

    unsubs.push(
      wsSchemaSubscription(data => {
        dispatch(setSchemaLoading(false));
        if (!data) {
          dispatch(addError('Failed to fetch schema'));
          return;
        } else {
          dispatch(readInSchemaFromBackend(data));
          dispatch(addInfo('Schema graph updated'));
        }
      }),
    );

    unsubs.push(
      wsQuerySubscription(data => {
        dispatch(setNewGraphQueryResult(data));
        dispatch(addInfo('Query Executed!'));
        dispatch(unSelect());
      }),
    );

    unsubs.push(
      wsTestSaveStateConnectionSubscription(data => {
        if (data?.status === 'success' && data.saveStateID) dispatch(testedSaveState(data.saveStateID));
        else {
          dispatch(addError('DB connection failed'));
        }
      }),
    );

    unsubs.push(
      wsReconnectSubscription(data => {
        if (data) dispatch(setSessionID(data));
      }),
    );

    Broker.instance().subscribe((data: SchemaGraphStats) => {
      dispatch(setSchemaAttributeInformation(data));
      dispatch(addInfo('Received attribute information'));
    }, wsReturnKey.schemaStatsResult);

    Broker.instance().subscribe((data: any) => {
      dispatch(setSchemaAttributeDimensions(JSON.parse(data)));
      dispatch(addInfo('Attribute dimensions added'));
    }, wsReturnKey.schemaInference);

    // Broker.instance().subscribe((data: QueryBuilderState) => dispatch(setQuerybuilderNodes(data)), 'query_builder_state');

    allMLTypes.forEach(mlType => {
      const id = Broker.instance().subscribe((data: LinkPredictionInstance[]) => {
        dispatch(setMLResult({ type: mlType, result: data }));
      }, mlType);
      unsubs.push(() => Broker.instance().unSubscribe(mlType, id));
    });

    unsubs.push(
      wsGetStatesSubscription((data, status) => {
        console.debug('Save States updated', data, status);
        if (status !== 'success' || !data) {
          dispatch(addError('Failed to fetch states'));
          return;
        }

        dispatch(updateSaveStateList(data));
        const d = Object.fromEntries(data.map(x => [x.id, x]));
        loadSaveState(session.currentSaveState, d);
        data.forEach(ss => {
          if (session.saveStatesAuthorization?.[ss.id] === undefined) {
            wsStateGetPolicy({ saveStateID: ss.id }, ret => {
              if (!ret) {
                dispatch(addError('Failed to fetch state authorization'));
                return;
              }
              dispatch(setStateAuthorization({ id: ss.id, authorization: ret }));
            });
          }
        });
      }),
    );

    unsubs.push(
      wsGetStateSubscription((data, status) => {
        if (status !== 'success' || !data) {
          dispatch(addError('Failed to fetch state'));
          return;
        }
        if (data.id !== nilUUID) {
          console.debug('Save State updated', data, status);
          dispatch(addSaveState(data));
          dispatch(selectSaveState(data.id));
          loadSaveState(data.id, session.saveStates);
          wsStateGetPolicy({ saveStateID: data.id }, ret => {
            if (!ret) {
              dispatch(addError('Failed to fetch state authorization'));
              return;
            }
            dispatch(setStateAuthorization({ id: data.id, authorization: ret }));
          });
        }
      }),
    );

    unsubs.push(wsDeleteStateSubscription(data => {}));

    // Broker.instance().subscribe((response: TestDatabaseConnectionResponse) => {
    //   if (response && response.status === 'success') dispatch(testedSaveState(response.saveStateID));
    // }, 'tested_connection');

    unsubs.push(
      wsQueryTranslationSubscription(response => {
        if (response && response.result) {
          dispatch(setQueryText(response));
        }
      }),
    );

    Broker.instance().subscribe((response: GraphQueryTranslationResultMessage) => {
      if (response && response.result) {
        dispatch(setQueryText(response));
      }
    }, wsReturnKey.queryStatusTranslationResult);

    login();

    // Setup cleanup
    return () => {
      // clear callback subscriptions
      unsubs.forEach(unsub => {
        unsub();
      });
      Broker.instance().unSubscribeAll(wsReturnKey.queryStatusTranslationResult);
      Broker.instance().unSubscribeAll(wsReturnKey.queryStatusResult);
      Broker.instance().unSubscribeAll(wsReturnKey.schemaInference);
    };
  }, []);

  useEffect(() => {
    if (session.currentSaveState) {
      const state = { ...session.saveStates[session.currentSaveState] };
      if (!isEqual(state.queries?.[0], queryBuilder)) {
        console.debug('Updating queryBuilder state', state.queries, queryBuilder);
        state.queries = [{ ...queryBuilder }];
        dispatch(queryingBackend());
        dispatch(updateSelectedSaveState(state));
        wsUpdateState(state);
      }
    }
  }, [queryBuilderSettings, queryHash, attributeShown]);

  useEffect(() => {
    if (session.currentSaveState) {
      const state = { ...session.saveStates[session.currentSaveState] };
      if (!isEqual(state.visualizations, visState)) {
        console.debug('Updating visState state', visState);
        state.visualizations = { ...visState };
        dispatch(updateSelectedSaveState(state));
        wsUpdateState(state);
      }
    }
  }, [visState]);

  useEffect(() => {
    // New active database
    if (session.currentSaveState && session.currentSaveState !== nilUUID) {
      dispatch(setSchemaLoading(true));
      wsSchemaRequest({ saveStateID: session.currentSaveState });
      wsSchemaStatsRequest({ saveStateID: session.currentSaveState });
      loadSaveState(session.currentSaveState, session.saveStates);
    }
  }, [session.currentSaveState]);

  useEffect(() => {
    // Newly (un)authorized
    if (auth.authentication?.authenticated && auth.authentication?.jwt) {
      props.onAuthorized();
      Broker.instance()
        .useAuth(auth.authentication)
        .connect(() => {
          console.debug('WS connected', session.currentSaveState, window.location.search);

          // Process URL Params
          const paramSaveState = getParam(URLParams.saveState);
          if (paramSaveState && paramSaveState !== nilUUID) {
            wsGetState({ saveStateID: paramSaveState });
          }

          dispatch(setFetchingSaveStates(true));

          // check authorizations
          wsUserGetPolicy(null, res => {
            if (res) dispatch(authorized(res));
          });

          wsGetStates(null, data => {
            if (data) {
              dispatch(setFetchingSaveStates(false));
              return true;
            } else {
              dispatch(addError('Failed to fetch states'));
              return false;
            }
          });

          // Broker.instance().sendMessage({ //TODO!!
          //   sessionID: auth?.sessionID || '',
          //   key: 'broadcastState',
          //   body: { type: 'subscribe', status: '', value: {} },
          // });
        });
    } else {
      // dispatch(logout());
    }
  }, [auth.authentication]);

  return <div className="hide"></div>;
};
